# generated by "generate_qnames"
from sdc11073.namespaces import default_ns_helper

AbstractAlertReport = default_ns_helper.MSG.tag('AbstractAlertReport')
AbstractComponentReport = default_ns_helper.MSG.tag('AbstractComponentReport')
AbstractContextReport = default_ns_helper.MSG.tag('AbstractContextReport')
AbstractGet = default_ns_helper.MSG.tag('AbstractGet')
AbstractGetResponse = default_ns_helper.MSG.tag('AbstractGetResponse')
AbstractMetricReport = default_ns_helper.MSG.tag('AbstractMetricReport')
AbstractOperationalStateReport = default_ns_helper.MSG.tag('AbstractOperationalStateReport')
AbstractReport = default_ns_helper.MSG.tag('AbstractReport')
AbstractReportPart = default_ns_helper.MSG.tag('AbstractReportPart')
AbstractSet = default_ns_helper.MSG.tag('AbstractSet')
AbstractSetResponse = default_ns_helper.MSG.tag('AbstractSetResponse')
Activate = default_ns_helper.MSG.tag('Activate')
ActivateResponse = default_ns_helper.MSG.tag('ActivateResponse')
AlertState = default_ns_helper.MSG.tag('AlertState')
ArgValue = default_ns_helper.MSG.tag('ArgValue')
Argument = default_ns_helper.MSG.tag('Argument')
By = default_ns_helper.MSG.tag('By')
ComponentState = default_ns_helper.MSG.tag('ComponentState')
ContainmentTree = default_ns_helper.MSG.tag('ContainmentTree')
ContextState = default_ns_helper.MSG.tag('ContextState')
DescriptionModificationReport = default_ns_helper.MSG.tag('DescriptionModificationReport')
Descriptor = default_ns_helper.MSG.tag('Descriptor')
DescriptorRevisions = default_ns_helper.MSG.tag('DescriptorRevisions')
EpisodicAlertReport = default_ns_helper.MSG.tag('EpisodicAlertReport')
EpisodicComponentReport = default_ns_helper.MSG.tag('EpisodicComponentReport')
EpisodicContextReport = default_ns_helper.MSG.tag('EpisodicContextReport')
EpisodicMetricReport = default_ns_helper.MSG.tag('EpisodicMetricReport')
EpisodicOperationalStateReport = default_ns_helper.MSG.tag('EpisodicOperationalStateReport')
ErrorCode = default_ns_helper.MSG.tag('ErrorCode')
ErrorInfo = default_ns_helper.MSG.tag('ErrorInfo')
Filter = default_ns_helper.MSG.tag('Filter')
GetContainmentTree = default_ns_helper.MSG.tag('GetContainmentTree')
GetContainmentTreeResponse = default_ns_helper.MSG.tag('GetContainmentTreeResponse')
GetContextStates = default_ns_helper.MSG.tag('GetContextStates')
GetContextStatesByFilter = default_ns_helper.MSG.tag('GetContextStatesByFilter')
GetContextStatesByFilterResponse = default_ns_helper.MSG.tag('GetContextStatesByFilterResponse')
GetContextStatesByIdentification = default_ns_helper.MSG.tag('GetContextStatesByIdentification')
GetContextStatesByIdentificationResponse = default_ns_helper.MSG.tag('GetContextStatesByIdentificationResponse')
GetContextStatesResponse = default_ns_helper.MSG.tag('GetContextStatesResponse')
GetDescriptor = default_ns_helper.MSG.tag('GetDescriptor')
GetDescriptorResponse = default_ns_helper.MSG.tag('GetDescriptorResponse')
GetDescriptorsFromArchive = default_ns_helper.MSG.tag('GetDescriptorsFromArchive')
GetDescriptorsFromArchiveResponse = default_ns_helper.MSG.tag('GetDescriptorsFromArchiveResponse')
GetLocalizedText = default_ns_helper.MSG.tag('GetLocalizedText')
GetLocalizedTextResponse = default_ns_helper.MSG.tag('GetLocalizedTextResponse')
GetMdDescription = default_ns_helper.MSG.tag('GetMdDescription')
GetMdDescriptionResponse = default_ns_helper.MSG.tag('GetMdDescriptionResponse')
GetMdState = default_ns_helper.MSG.tag('GetMdState')
GetMdStateResponse = default_ns_helper.MSG.tag('GetMdStateResponse')
GetMdib = default_ns_helper.MSG.tag('GetMdib')
GetMdibResponse = default_ns_helper.MSG.tag('GetMdibResponse')
GetStatesFromArchive = default_ns_helper.MSG.tag('GetStatesFromArchive')
GetStatesFromArchiveResponse = default_ns_helper.MSG.tag('GetStatesFromArchiveResponse')
GetSupportedLanguages = default_ns_helper.MSG.tag('GetSupportedLanguages')
GetSupportedLanguagesResponse = default_ns_helper.MSG.tag('GetSupportedLanguagesResponse')
Handle = default_ns_helper.MSG.tag('Handle')
HandleRef = default_ns_helper.MSG.tag('HandleRef')
Identification = default_ns_helper.MSG.tag('Identification')
InvocationError = default_ns_helper.MSG.tag('InvocationError')
InvocationErrorMessage = default_ns_helper.MSG.tag('InvocationErrorMessage')
InvocationInfo = default_ns_helper.MSG.tag('InvocationInfo')
InvocationSource = default_ns_helper.MSG.tag('InvocationSource')
InvocationState = default_ns_helper.MSG.tag('InvocationState')
Lang = default_ns_helper.MSG.tag('Lang')
MdDescription = default_ns_helper.MSG.tag('MdDescription')
MdState = default_ns_helper.MSG.tag('MdState')
Mdib = default_ns_helper.MSG.tag('Mdib')
MetricState = default_ns_helper.MSG.tag('MetricState')
NumberOfLines = default_ns_helper.MSG.tag('NumberOfLines')
ObservedValueStream = default_ns_helper.MSG.tag('ObservedValueStream')
OperationHandleRef = default_ns_helper.MSG.tag('OperationHandleRef')
OperationInvokedReport = default_ns_helper.MSG.tag('OperationInvokedReport')
OperationState = default_ns_helper.MSG.tag('OperationState')
PeriodicAlertReport = default_ns_helper.MSG.tag('PeriodicAlertReport')
PeriodicComponentReport = default_ns_helper.MSG.tag('PeriodicComponentReport')
PeriodicContextReport = default_ns_helper.MSG.tag('PeriodicContextReport')
PeriodicMetricReport = default_ns_helper.MSG.tag('PeriodicMetricReport')
PeriodicOperationalStateReport = default_ns_helper.MSG.tag('PeriodicOperationalStateReport')
ProposedAlertState = default_ns_helper.MSG.tag('ProposedAlertState')
ProposedComponentState = default_ns_helper.MSG.tag('ProposedComponentState')
ProposedContextState = default_ns_helper.MSG.tag('ProposedContextState')
ProposedMetricState = default_ns_helper.MSG.tag('ProposedMetricState')
Ref = default_ns_helper.MSG.tag('Ref')
ReportPart = default_ns_helper.MSG.tag('ReportPart')
RequestedNumericValue = default_ns_helper.MSG.tag('RequestedNumericValue')
RequestedStringValue = default_ns_helper.MSG.tag('RequestedStringValue')
Retrievability = default_ns_helper.MSG.tag('Retrievability')
RetrievabilityInfo = default_ns_helper.MSG.tag('RetrievabilityInfo')
SetAlertState = default_ns_helper.MSG.tag('SetAlertState')
SetAlertStateResponse = default_ns_helper.MSG.tag('SetAlertStateResponse')
SetComponentState = default_ns_helper.MSG.tag('SetComponentState')
SetComponentStateResponse = default_ns_helper.MSG.tag('SetComponentStateResponse')
SetContextState = default_ns_helper.MSG.tag('SetContextState')
SetContextStateResponse = default_ns_helper.MSG.tag('SetContextStateResponse')
SetMetricState = default_ns_helper.MSG.tag('SetMetricState')
SetMetricStateResponse = default_ns_helper.MSG.tag('SetMetricStateResponse')
SetString = default_ns_helper.MSG.tag('SetString')
SetStringResponse = default_ns_helper.MSG.tag('SetStringResponse')
SetValue = default_ns_helper.MSG.tag('SetValue')
SetValueResponse = default_ns_helper.MSG.tag('SetValueResponse')
SourceMds = default_ns_helper.MSG.tag('SourceMds')
State = default_ns_helper.MSG.tag('State')
StateRevisions = default_ns_helper.MSG.tag('StateRevisions')
SystemErrorReport = default_ns_helper.MSG.tag('SystemErrorReport')
Text = default_ns_helper.MSG.tag('Text')
TextWidth = default_ns_helper.MSG.tag('TextWidth')
TimeFrame = default_ns_helper.MSG.tag('TimeFrame')
TransactionId = default_ns_helper.MSG.tag('TransactionId')
Value = default_ns_helper.MSG.tag('Value')
Version = default_ns_helper.MSG.tag('Version')
VersionFrame = default_ns_helper.MSG.tag('VersionFrame')
WaveformStream = default_ns_helper.MSG.tag('WaveformStream')
