# generated by "generate_qnames"
from sdc11073.namespaces import default_ns_helper

AbstractAlertDescriptor = default_ns_helper.PM.tag('AbstractAlertDescriptor')
AbstractAlertState = default_ns_helper.PM.tag('AbstractAlertState')
AbstractComplexDeviceComponentDescriptor = default_ns_helper.PM.tag('AbstractComplexDeviceComponentDescriptor')
AbstractComplexDeviceComponentState = default_ns_helper.PM.tag('AbstractComplexDeviceComponentState')
AbstractContextDescriptor = default_ns_helper.PM.tag('AbstractContextDescriptor')
AbstractContextState = default_ns_helper.PM.tag('AbstractContextState')
AbstractDescriptor = default_ns_helper.PM.tag('AbstractDescriptor')
AbstractDeviceComponentDescriptor = default_ns_helper.PM.tag('AbstractDeviceComponentDescriptor')
AbstractDeviceComponentState = default_ns_helper.PM.tag('AbstractDeviceComponentState')
AbstractMetricDescriptor = default_ns_helper.PM.tag('AbstractMetricDescriptor')
AbstractMetricState = default_ns_helper.PM.tag('AbstractMetricState')
AbstractMetricValue = default_ns_helper.PM.tag('AbstractMetricValue')
AbstractMultiState = default_ns_helper.PM.tag('AbstractMultiState')
AbstractOperationDescriptor = default_ns_helper.PM.tag('AbstractOperationDescriptor')
AbstractOperationState = default_ns_helper.PM.tag('AbstractOperationState')
AbstractSetStateOperationDescriptor = default_ns_helper.PM.tag('AbstractSetStateOperationDescriptor')
AbstractState = default_ns_helper.PM.tag('AbstractState')
AccessionIdentifier = default_ns_helper.PM.tag('AccessionIdentifier')
ActivateOperationDescriptor = default_ns_helper.PM.tag('ActivateOperationDescriptor')
ActivateOperationState = default_ns_helper.PM.tag('ActivateOperationState')
ActiveSyncProtocol = default_ns_helper.PM.tag('ActiveSyncProtocol')
AlertCondition = default_ns_helper.PM.tag('AlertCondition')
AlertConditionDescriptor = default_ns_helper.PM.tag('AlertConditionDescriptor')
AlertConditionState = default_ns_helper.PM.tag('AlertConditionState')
AlertSignal = default_ns_helper.PM.tag('AlertSignal')
AlertSignalDescriptor = default_ns_helper.PM.tag('AlertSignalDescriptor')
AlertSignalState = default_ns_helper.PM.tag('AlertSignalState')
AlertSystem = default_ns_helper.PM.tag('AlertSystem')
AlertSystemDescriptor = default_ns_helper.PM.tag('AlertSystemDescriptor')
AlertSystemState = default_ns_helper.PM.tag('AlertSystemState')
AllowedRange = default_ns_helper.PM.tag('AllowedRange')
AllowedValue = default_ns_helper.PM.tag('AllowedValue')
AllowedValues = default_ns_helper.PM.tag('AllowedValues')
Annotation = default_ns_helper.PM.tag('Annotation')
ApplyAnnotation = default_ns_helper.PM.tag('ApplyAnnotation')
ApprovedJurisdiction = default_ns_helper.PM.tag('ApprovedJurisdiction')
ApprovedJurisdictions = default_ns_helper.PM.tag('ApprovedJurisdictions')
Arg = default_ns_helper.PM.tag('Arg')
ArgName = default_ns_helper.PM.tag('ArgName')
Argument = default_ns_helper.PM.tag('Argument')
AssignedLocation = default_ns_helper.PM.tag('AssignedLocation')
BaseDemographics = default_ns_helper.PM.tag('BaseDemographics')
Battery = default_ns_helper.PM.tag('Battery')
BatteryDescriptor = default_ns_helper.PM.tag('BatteryDescriptor')
BatteryState = default_ns_helper.PM.tag('BatteryState')
BirthLength = default_ns_helper.PM.tag('BirthLength')
BirthWeight = default_ns_helper.PM.tag('BirthWeight')
Birthname = default_ns_helper.PM.tag('Birthname')
BodySite = default_ns_helper.PM.tag('BodySite')
CalibrationDocumentation = default_ns_helper.PM.tag('CalibrationDocumentation')
CalibrationInfo = default_ns_helper.PM.tag('CalibrationInfo')
CalibrationResult = default_ns_helper.PM.tag('CalibrationResult')
CapacityFullCharge = default_ns_helper.PM.tag('CapacityFullCharge')
CapacityRemaining = default_ns_helper.PM.tag('CapacityRemaining')
CapacitySpecified = default_ns_helper.PM.tag('CapacitySpecified')
Category = default_ns_helper.PM.tag('Category')
CauseInfo = default_ns_helper.PM.tag('CauseInfo')
Channel = default_ns_helper.PM.tag('Channel')
ChannelDescriptor = default_ns_helper.PM.tag('ChannelDescriptor')
ChannelState = default_ns_helper.PM.tag('ChannelState')
Characteristic = default_ns_helper.PM.tag('Characteristic')
ClinicalInfo = default_ns_helper.PM.tag('ClinicalInfo')
Clock = default_ns_helper.PM.tag('Clock')
ClockDescriptor = default_ns_helper.PM.tag('ClockDescriptor')
ClockState = default_ns_helper.PM.tag('ClockState')
Code = default_ns_helper.PM.tag('Code')
CodedValue = default_ns_helper.PM.tag('CodedValue')
CodingSystemName = default_ns_helper.PM.tag('CodingSystemName')
ComponentId = default_ns_helper.PM.tag('ComponentId')
ConceptDescription = default_ns_helper.PM.tag('ConceptDescription')
ContainmentTree = default_ns_helper.PM.tag('ContainmentTree')
ContainmentTreeEntry = default_ns_helper.PM.tag('ContainmentTreeEntry')
CoreData = default_ns_helper.PM.tag('CoreData')
Criticality = default_ns_helper.PM.tag('Criticality')
Current = default_ns_helper.PM.tag('Current')
DangerCode = default_ns_helper.PM.tag('DangerCode')
DateOfBirth = default_ns_helper.PM.tag('DateOfBirth')
Description = default_ns_helper.PM.tag('Description')
DeviceIdentifier = default_ns_helper.PM.tag('DeviceIdentifier')
DistributionRange = default_ns_helper.PM.tag('DistributionRange')
DistributionSampleArrayMetricDescriptor = default_ns_helper.PM.tag('DistributionSampleArrayMetricDescriptor')
DistributionSampleArrayMetricState = default_ns_helper.PM.tag('DistributionSampleArrayMetricState')
Documentation = default_ns_helper.PM.tag('Documentation')
DomainUnit = default_ns_helper.PM.tag('DomainUnit')
End = default_ns_helper.PM.tag('End')
EnsembleContext = default_ns_helper.PM.tag('EnsembleContext')
EnsembleContextDescriptor = default_ns_helper.PM.tag('EnsembleContextDescriptor')
EnsembleContextState = default_ns_helper.PM.tag('EnsembleContextState')
Entry = default_ns_helper.PM.tag('Entry')
EnumStringMetricDescriptor = default_ns_helper.PM.tag('EnumStringMetricDescriptor')
EnumStringMetricState = default_ns_helper.PM.tag('EnumStringMetricState')
ExpirationDate = default_ns_helper.PM.tag('ExpirationDate')
Familyname = default_ns_helper.PM.tag('Familyname')
FillerOrderNumber = default_ns_helper.PM.tag('FillerOrderNumber')
GestationalAge = default_ns_helper.PM.tag('GestationalAge')
Givenname = default_ns_helper.PM.tag('Givenname')
HeadCircumference = default_ns_helper.PM.tag('HeadCircumference')
Height = default_ns_helper.PM.tag('Height')
HumanReadableForm = default_ns_helper.PM.tag('HumanReadableForm')
Identification = default_ns_helper.PM.tag('Identification')
IdentifierName = default_ns_helper.PM.tag('IdentifierName')
ImagingProcedure = default_ns_helper.PM.tag('ImagingProcedure')
InstanceIdentifier = default_ns_helper.PM.tag('InstanceIdentifier')
Issuer = default_ns_helper.PM.tag('Issuer')
Jurisdiction = default_ns_helper.PM.tag('Jurisdiction')
Label = default_ns_helper.PM.tag('Label')
LimitAlertConditionDescriptor = default_ns_helper.PM.tag('LimitAlertConditionDescriptor')
LimitAlertConditionState = default_ns_helper.PM.tag('LimitAlertConditionState')
Limits = default_ns_helper.PM.tag('Limits')
LocalizedText = default_ns_helper.PM.tag('LocalizedText')
LocationContext = default_ns_helper.PM.tag('LocationContext')
LocationContextDescriptor = default_ns_helper.PM.tag('LocationContextDescriptor')
LocationContextState = default_ns_helper.PM.tag('LocationContextState')
LocationDetail = default_ns_helper.PM.tag('LocationDetail')
LocationReference = default_ns_helper.PM.tag('LocationReference')
LotNumber = default_ns_helper.PM.tag('LotNumber')
ManufactureDate = default_ns_helper.PM.tag('ManufactureDate')
Manufacturer = default_ns_helper.PM.tag('Manufacturer')
MaxLimits = default_ns_helper.PM.tag('MaxLimits')
MdDescription = default_ns_helper.PM.tag('MdDescription')
MdState = default_ns_helper.PM.tag('MdState')
Mdib = default_ns_helper.PM.tag('Mdib')
Mds = default_ns_helper.PM.tag('Mds')
MdsDescriptor = default_ns_helper.PM.tag('MdsDescriptor')
MdsState = default_ns_helper.PM.tag('MdsState')
Meaning = default_ns_helper.PM.tag('Meaning')
MeansContext = default_ns_helper.PM.tag('MeansContext')
MeansContextDescriptor = default_ns_helper.PM.tag('MeansContextDescriptor')
MeansContextState = default_ns_helper.PM.tag('MeansContextState')
Measurement = default_ns_helper.PM.tag('Measurement')
MeasurementUnit = default_ns_helper.PM.tag('MeasurementUnit')
MetaData = default_ns_helper.PM.tag('MetaData')
Metric = default_ns_helper.PM.tag('Metric')
MetricQuality = default_ns_helper.PM.tag('MetricQuality')
MetricValue = default_ns_helper.PM.tag('MetricValue')
Middlename = default_ns_helper.PM.tag('Middlename')
Modality = default_ns_helper.PM.tag('Modality')
ModelName = default_ns_helper.PM.tag('ModelName')
ModelNumber = default_ns_helper.PM.tag('ModelNumber')
ModifiableData = default_ns_helper.PM.tag('ModifiableData')
Mother = default_ns_helper.PM.tag('Mother')
Name = default_ns_helper.PM.tag('Name')
NeonatalPatientDemographicsCoreData = default_ns_helper.PM.tag('NeonatalPatientDemographicsCoreData')
NextCalibration = default_ns_helper.PM.tag('NextCalibration')
NumericMetricDescriptor = default_ns_helper.PM.tag('NumericMetricDescriptor')
NumericMetricState = default_ns_helper.PM.tag('NumericMetricState')
NumericMetricValue = default_ns_helper.PM.tag('NumericMetricValue')
OperatingJurisdiction = default_ns_helper.PM.tag('OperatingJurisdiction')
Operation = default_ns_helper.PM.tag('Operation')
OperationGroup = default_ns_helper.PM.tag('OperationGroup')
OperatorContext = default_ns_helper.PM.tag('OperatorContext')
OperatorContextDescriptor = default_ns_helper.PM.tag('OperatorContextDescriptor')
OperatorContextState = default_ns_helper.PM.tag('OperatorContextState')
OperatorDetails = default_ns_helper.PM.tag('OperatorDetails')
OrderDetail = default_ns_helper.PM.tag('OrderDetail')
Patient = default_ns_helper.PM.tag('Patient')
PatientContext = default_ns_helper.PM.tag('PatientContext')
PatientContextDescriptor = default_ns_helper.PM.tag('PatientContextDescriptor')
PatientContextState = default_ns_helper.PM.tag('PatientContextState')
PatientDemographicsCoreData = default_ns_helper.PM.tag('PatientDemographicsCoreData')
PatientType = default_ns_helper.PM.tag('PatientType')
PerformedOrderDetail = default_ns_helper.PM.tag('PerformedOrderDetail')
Performer = default_ns_helper.PM.tag('Performer')
PersonParticipation = default_ns_helper.PM.tag('PersonParticipation')
PersonReference = default_ns_helper.PM.tag('PersonReference')
PhysicalConnector = default_ns_helper.PM.tag('PhysicalConnector')
PhysicalConnectorInfo = default_ns_helper.PM.tag('PhysicalConnectorInfo')
PhysiologicalRange = default_ns_helper.PM.tag('PhysiologicalRange')
PlacerOrderNumber = default_ns_helper.PM.tag('PlacerOrderNumber')
ProductionSpec = default_ns_helper.PM.tag('ProductionSpec')
ProductionSpecification = default_ns_helper.PM.tag('ProductionSpecification')
ProtocolCode = default_ns_helper.PM.tag('ProtocolCode')
Race = default_ns_helper.PM.tag('Race')
Range = default_ns_helper.PM.tag('Range')
RealTimeSampleArrayMetricDescriptor = default_ns_helper.PM.tag('RealTimeSampleArrayMetricDescriptor')
RealTimeSampleArrayMetricState = default_ns_helper.PM.tag('RealTimeSampleArrayMetricState')
ReferenceRange = default_ns_helper.PM.tag('ReferenceRange')
ReferenceSource = default_ns_helper.PM.tag('ReferenceSource')
ReferringPhysician = default_ns_helper.PM.tag('ReferringPhysician')
RelatedMeasurement = default_ns_helper.PM.tag('RelatedMeasurement')
Relation = default_ns_helper.PM.tag('Relation')
RelevantClinicalInfo = default_ns_helper.PM.tag('RelevantClinicalInfo')
RemainingBatteryTime = default_ns_helper.PM.tag('RemainingBatteryTime')
RemedyInfo = default_ns_helper.PM.tag('RemedyInfo')
RequestedOrderDetail = default_ns_helper.PM.tag('RequestedOrderDetail')
RequestedProcedureId = default_ns_helper.PM.tag('RequestedProcedureId')
RequestingPhysician = default_ns_helper.PM.tag('RequestingPhysician')
ResultingClinicalInfo = default_ns_helper.PM.tag('ResultingClinicalInfo')
Role = default_ns_helper.PM.tag('Role')
SampleArrayValue = default_ns_helper.PM.tag('SampleArrayValue')
ScheduledProcedureStepId = default_ns_helper.PM.tag('ScheduledProcedureStepId')
Sco = default_ns_helper.PM.tag('Sco')
ScoDescriptor = default_ns_helper.PM.tag('ScoDescriptor')
ScoState = default_ns_helper.PM.tag('ScoState')
SerialNumber = default_ns_helper.PM.tag('SerialNumber')
Service = default_ns_helper.PM.tag('Service')
SetAlertStateOperationDescriptor = default_ns_helper.PM.tag('SetAlertStateOperationDescriptor')
SetAlertStateOperationState = default_ns_helper.PM.tag('SetAlertStateOperationState')
SetComponentStateOperationDescriptor = default_ns_helper.PM.tag('SetComponentStateOperationDescriptor')
SetComponentStateOperationState = default_ns_helper.PM.tag('SetComponentStateOperationState')
SetContextStateOperationDescriptor = default_ns_helper.PM.tag('SetContextStateOperationDescriptor')
SetContextStateOperationState = default_ns_helper.PM.tag('SetContextStateOperationState')
SetMetricStateOperationDescriptor = default_ns_helper.PM.tag('SetMetricStateOperationDescriptor')
SetMetricStateOperationState = default_ns_helper.PM.tag('SetMetricStateOperationState')
SetStringOperationDescriptor = default_ns_helper.PM.tag('SetStringOperationDescriptor')
SetStringOperationState = default_ns_helper.PM.tag('SetStringOperationState')
SetValueOperationDescriptor = default_ns_helper.PM.tag('SetValueOperationDescriptor')
SetValueOperationState = default_ns_helper.PM.tag('SetValueOperationState')
Sex = default_ns_helper.PM.tag('Sex')
Source = default_ns_helper.PM.tag('Source')
SpecType = default_ns_helper.PM.tag('SpecType')
Start = default_ns_helper.PM.tag('Start')
State = default_ns_helper.PM.tag('State')
StringMetricDescriptor = default_ns_helper.PM.tag('StringMetricDescriptor')
StringMetricState = default_ns_helper.PM.tag('StringMetricState')
StringMetricValue = default_ns_helper.PM.tag('StringMetricValue')
StudyInstanceUid = default_ns_helper.PM.tag('StudyInstanceUid')
SystemContext = default_ns_helper.PM.tag('SystemContext')
SystemContextDescriptor = default_ns_helper.PM.tag('SystemContextDescriptor')
SystemContextState = default_ns_helper.PM.tag('SystemContextState')
SystemSignalActivation = default_ns_helper.PM.tag('SystemSignalActivation')
TechnicalRange = default_ns_helper.PM.tag('TechnicalRange')
Temperature = default_ns_helper.PM.tag('Temperature')
TimeProtocol = default_ns_helper.PM.tag('TimeProtocol')
Title = default_ns_helper.PM.tag('Title')
Translation = default_ns_helper.PM.tag('Translation')
Type = default_ns_helper.PM.tag('Type')
Udi = default_ns_helper.PM.tag('Udi')
Unit = default_ns_helper.PM.tag('Unit')
Validator = default_ns_helper.PM.tag('Validator')
Value = default_ns_helper.PM.tag('Value')
VisitNumber = default_ns_helper.PM.tag('VisitNumber')
Vmd = default_ns_helper.PM.tag('Vmd')
VmdDescriptor = default_ns_helper.PM.tag('VmdDescriptor')
VmdState = default_ns_helper.PM.tag('VmdState')
Voltage = default_ns_helper.PM.tag('Voltage')
VoltageSpecified = default_ns_helper.PM.tag('VoltageSpecified')
Weight = default_ns_helper.PM.tag('Weight')
WorkflowContext = default_ns_helper.PM.tag('WorkflowContext')
WorkflowContextDescriptor = default_ns_helper.PM.tag('WorkflowContextDescriptor')
WorkflowContextState = default_ns_helper.PM.tag('WorkflowContextState')
WorkflowDetail = default_ns_helper.PM.tag('WorkflowDetail')
